/* --------------------------------------------------------------------------
 * Copyright 1992 by Forschungszentrum Informatik (FZI)
 *
 * You can use and distribute this software under the terms of the licence
 * you should have received along with this program.
 * If not or if you want additional information, write to
 * Forschungszentrum Informatik, "STONE", Haid-und-Neu-Strasse 10-14,
 * D-7500 Karlsruhe 1, Germany.
 * --------------------------------------------------------------------------
 */
// **************************************************************************
// Module dmp_main                                           Juergen Uhl (ju)
//
// **************************************************************************
// SOS dump tool
// **************************************************************************

#include <stdio.h>
#include <string.h>
#include "sys.h"
#include "sos.h"
#include "dmp_err.h"
#include "mta_use.h"
#include "dir_use.h"
#include "cci_use.h"

static sos_Object_Set dumped_objects;
static sos_Bool only_one_container;
static sos_Bool only_container;
static sos_Container root_container;

static void dump_object (sos_Object o, sos_Int level,
                         sos_Bool as_component = FALSE);

sos_Object get_component (sos_Object o, sos_Int os, sos_Type tp,
                          sos_Bool is_local)
{  sos_Typed_id tpid;

   if (tp.is_scalar ())
   {  sos_Int sz = tp.get_object_size();
      sos_Id id = sos_Id::make (sos_Container::make (0), 0);
      o.container().read (o.offset() + os,sz,&id);
      tpid = sos_Typed_id::make (id, tp.typed_id().get_id());
   }
   else if (is_local)
   {  sos_Char c[SOS_OFFSET_SIZE];
      o.container().read (o.offset() + os,SOS_OFFSET_SIZE,c);
      sos_Offset os1;
      bcopy_to_sos_Offset(&os1,c);
      tpid = sos_make_local_typed_id (os1, o.container());
   }
   else
   {  sos_Char c[SOS_TYPED_ID_SIZE];
      o.container().read (o.offset() + os,SOS_TYPED_ID_SIZE,c);
      bcopy_to_sos_Typed_id(&tpid,c);
   }

   return sos_Object::make (tpid);
}

void dump_immediate_components (sos_Object o, sos_Class_type ct, sos_Int os,
				sos_Int level, sos_Bool as_component,
				sos_Bool &is_first)
{  sos_Method_List ml = ct.get_comp_methods();
   agg_iterate (ml, sos_Method m)
   {  sos_Comp_method cm = sos_Comp_method::make (m);
      if (NOT cm.get_is_set ())
      {  sos_Type ctp = cm.get_result_type().make_base_type();
	 if (NOT (only_container AND ctp.is_scalar()))
	 {  sos_Object co = get_component (o, os + cm.get_offset(),
					   ctp, cm.get_is_local());
	    if (as_component)
	    {  if (is_first)
		  is_first = FALSE;
	       else
		  printf (",");
	       char *cn = cm.get_name().make_Cstring();
	       printf ("%s=", cn + 4); // +4 skips "get_"
	       delete cn;
	    }
	    dump_object (co, (as_component?level:level-1), as_component);
	 }
      }
   }
   agg_iterate_end (ml, m);

#ifdef ATT
   if (as_component AND ct.operator==(sos_String_type))
#else
   if (as_component AND ct == sos_String_type)
#endif
   {  if (NOT is_first) printf (",");
      char *s = sos_String::make (o).make_Cstring();
      printf ("*value*=\"%s\"", s);
      delete s;
   }
#ifdef ATT
   else if (ct.root_class().operator==(sos_Object_Collection_type))
#else
   else if (ct.root_class() == sos_Object_Collection_type)
#endif
   {  if (as_component)
      {  if  (is_first)
	   is_first = FALSE;
	 else
	    printf (",");
	 printf ("*value*={");
      }
      sos_Bool is_first_elem = TRUE;
      sos_Object_Collection coll = sos_Object_Collection::make (o);
      agg_iterate (coll, sos_Object co)
      {  if (as_component)
	 {  if (is_first_elem)
	       is_first_elem = FALSE;
	    else
	       printf (",");
	 }
	 dump_object (co, (as_component?level:level-1), as_component);
      }
      agg_iterate_end (coll, co);
      if (as_component) printf ("}");
   }
#ifdef ATT
   else if (ct.root_class().operator==(sos_Object_sos_Object_Association_type))
#else
   else if (ct.root_class() == sos_Object_sos_Object_Association_type)
#endif
   {  if (as_component)
      {  if  (is_first)
	   is_first = FALSE;
	 else
	    printf (",");
	 printf ("*value*={");
      }
      sos_Bool is_first_elem = TRUE;
      sos_Object_sos_Object_Association as =
	 sos_Object_sos_Object_Association::make (o);
      agg_iterate_association (as, sos_Object role1, sos_Object role2)
      {  if (as_component)
	 {  if (is_first_elem)
	       is_first_elem = FALSE;
	    else
	       printf (",");
	    printf ("<");
	 }
	 dump_object (role1, (as_component?level:level-1), as_component);
         if (as_component) printf (",");
	 dump_object (role2, (as_component?level:level-1), as_component);
         if (as_component) printf (">");
      }
      agg_iterate_association_end (as, role1, role2);
      if (as_component) printf ("}");
   }
}

void dump_class_instance (sos_Object o, sos_Class_type ct,
			  sos_Int level, sos_Bool as_component)
{  sos_Super_class_List scl = ct.root_class().get_super_closure();
   sos_Bool is_first = TRUE;
   agg_iterate (scl, sos_Super_class sc)
   {  sos_Class_type sct = sc.get_super_class();
      sos_Int os = sc.get_offset();
      dump_immediate_components (o, sct, os, level, as_component, is_first);
   }
   agg_iterate_end (scl, sc)
}

static sos_Object_Set dumped_containers;

void dump_container (sos_Container ct)
{  sos_Object oct = make_sos_Container_object (ct);
   if (dumped_containers.is_element (oct)) return;
   dumped_containers.insert (oct);
   printf ("%d\n", sos_Int (ct));
}

void dump_object (sos_Object o, sos_Int level, sos_Bool as_component /* = FALSE */ )
{  if (level == 0) return;

   if (o == NO_OBJECT)
   {  if (as_component) printf ("<>");
      return;
   }
   if (NOT as_component AND
       only_one_container AND
       root_container != o.container())
      return;

   sos_Type tp = o.type();

   if (as_component)
   {  err_block
      {  if (tp.has_type (sos_Class_type_type))
	    printf ("<%d,%d>", (sos_Int) o.container(), o.offset());
	 else if (tp.has_type (sos_Enum_type_type))
	 {  sos_Int     i  = sos_Int_from_enum (o, tp);
	    sos_Cstring ln = sos_Enum_type::make (tp).
			     get_literals().get_nth(i+1).make_Cstring();
	    printf ("%s:", ln);
	    delete ln;
	 }
	 else if (tp.has_type (sos_Extern_type_type))
	 {  sos_Cstring s = cci_Schema_impl::string_from_extern_object (o).
			    make_Cstring();
	    printf (s);
	    delete s;
	 }
	 else
	    printf ("???");
      }
      err_exception
      {  printf ("???");
      }
      err_block_end
   }
   else if (tp.has_type (sos_Class_type_type))
   {  if (dumped_objects.is_element (o)) return;
      dumped_objects.insert (o);
      if (only_container)
	 dump_container (o.container());
      else
      {  err_block
	 {  printf ("<%d/%d>:", (sos_Int) o.container(), o.offset());
	    char *tn = tp.get_name().make_Cstring();
	    printf ("[%s(%d):", tn, o.size());
	    delete tn;
	    dump_class_instance (o, sos_Class_type::make (tp), level, TRUE);
	    printf ("]\n");
	 }
	 err_exception
	 {  printf ("???\n");
	 }
	 err_block_end
      }
      dump_class_instance (o, sos_Class_type::make (tp), level, FALSE);
   }
}

#ifdef PROFILE
extern "C" void monitor(...);
#endif

main (int argc, char *argv[])
{  
   sos_Object o;
   sos_Int level=-1;
   sos_Int i=1;

   sos_init (argc, argv);

   only_one_container = FALSE;
   only_container = FALSE;

   while (argc > i AND argv[i][0] == '-')
   {  if (argc > i+1 AND streql (argv [i], "-l"))
      {  level = atoi (argv [i+1]);
	 if (level==0)
	 {  err_raise (err_USE, err_DMP_USAGE, NULL, FALSE);
	    exit (-1);
	 }
	 i+=2;
      }
      else if (streql (argv [i], "-c"))
      {  only_container = TRUE;
	 i++;
      }
      else if (streql (argv [i], "-o"))
      {  only_one_container = TRUE;
	 i++;
      }
      else
      {  err_raise (err_USE, err_DMP_USAGE, NULL, FALSE);
	 exit (-1);
      }
   }

   if (argc > i+2)
   {  err_raise (err_USE, err_DMP_USAGE, NULL, FALSE);
      exit (-1);
   }
   else if (argc > i)
   {  sos_Container ct;
      sos_Offset os;
      ct = sos_Container::make (atoi (argv [i]));
      if (argc > i+1)
	 os = atoi (argv [i+1]);
      else
	 os = ROOT_OFFSET;
      if (ct == 0 OR os == 0)
      {  err_raise (err_USE, err_DMP_USAGE, NULL, FALSE);
	 exit (-1);
      }
      err_block
      {  o = sos_Object::make
		(sos_Typed_id::make
		   (sos_Id::make (sos_Container::make (ct), os)));
      }
      err_exception
      {  err_raise (err_USE, err_DMP_INVALID_ROOT, NULL, FALSE);
	 exit (-1);
      }
      err_block_end
   }
   else
      o = sos_Object_Directory::root();

   sos_Container dct = sos_Container::create();
   dumped_objects = sos_Object_Set::create (dct, FALSE);
   dumped_containers = sos_Object_Set::create (dct, FALSE);
   root_container = o.container();
   dump_object (o, level);
   dct.destroy();
   dct.close();

#ifdef PROFILE
   monitor (0);
#endif

   exit (0);
}
